@@ -31,7 +31,7 @@
 -include("ssl_internal.hrl").
 -include_lib("public_key/include/public_key.hrl"). 
 
--export([trusted_cert_and_path/4,
+-export([trusted_cert_and_paths/4,
 	 certificate_chain/3,
 	 certificate_chain/4,
 	 file_to_certificats/2,
@@ -50,46 +50,46 @@
 %%====================================================================
 
 %%--------------------------------------------------------------------
--spec trusted_cert_and_path([der_cert()], db_handle(), certdb_ref(), fun()) ->
-				   {der_cert() | unknown_ca, [der_cert()]}.
+-spec trusted_cert_and_paths([der_cert()], db_handle(), certdb_ref(), fun()) ->
+				   [{der_cert() | unknown_ca | invalid_issuer | selfsigned_peer, [der_cert()]}].
 %%
-%% Description: Extracts the root cert (if not presents tries to 
-%% look it up, if not found {bad_cert, unknown_ca} will be added verification
-%% errors. Returns {RootCert, Path, VerifyErrors}
+%% Description: Construct input to public_key:pkix_path_validation/3,
+%% If the ROOT cert is not found {bad_cert, unknown_ca} will be returned
+%% instead of the ROOT cert to be handled as a path validation error
+%% by the verify_fun. 
+%% Returns {RootCert | RootCertRelatedError, Path} 
+%% Note: Path = lists:reverse(Chain) -- Root, that is on the peer cert 
+%% always comes first in the chain but last in the path.
 %%--------------------------------------------------------------------
-trusted_cert_and_path(CertChain, CertDbHandle, CertDbRef, PartialChainHandler) ->
-    Path = [BinCert | _] = lists:reverse(CertChain),
-    OtpCert = public_key:pkix_decode_cert(BinCert, otp),
-    SignedAndIssuerID =
-	case public_key:pkix_is_self_signed(OtpCert) of
-	    true ->
-		{ok, IssuerId} = public_key:pkix_issuer_id(OtpCert, self),
-		{self, IssuerId};
-	    false ->
-		other_issuer(OtpCert, BinCert, CertDbHandle, CertDbRef)
-	end,
-    
-    case SignedAndIssuerID of
-	{error, issuer_not_found} ->
-	    %% The root CA was not sent and cannot be found.
-	    handle_incomplete_chain(Path, PartialChainHandler);
-	{self, _} when length(Path) == 1 ->
-	    {selfsigned_peer, Path};
-	{_ ,{SerialNr, Issuer}} ->
-	    case ssl_manager:lookup_trusted_cert(CertDbHandle, CertDbRef, SerialNr, Issuer) of
-		{ok, Trusted} ->
-		    %% Trusted must be selfsigned or it is an incomplete chain
-		    handle_path(Trusted, Path, PartialChainHandler);
-		_ ->
-		    %% Root CA could not be verified, but partial
-                    %% chain handler may trusted a cert that we got
-		    handle_incomplete_chain(Path, PartialChainHandler)
-	    end
-    end.
-
+trusted_cert_and_paths([Peer] = Chain,  CertDbHandle, CertDbRef, PartialChainHandler) ->
+    OtpCert = public_key:pkix_decode_cert(Peer, otp),
+    case public_key:pkix_is_self_signed(OtpCert) of
+        true ->
+            [{selfsigned_peer, [Peer]}];
+        false ->
+            [handle_incomplete_chain(Chain, PartialChainHandler, {unknown_ca, [Peer]},
+                                     CertDbHandle, CertDbRef)]
+    end;
+trusted_cert_and_paths(Chain,  CertDbHandle, CertDbRef, PartialChainHandler) ->
+    %% Construct possible certificate paths from the chain certificates.
+    %% If the chain contains extraneous certificates there could be
+    %% more than one possible path such chains might be used to phase out 
+    %% an old certificate. 
+    Paths = paths(Chain, CertDbHandle),
+    lists:map(fun(Path) ->
+                      case handle_partial_chain(Path, PartialChainHandler, CertDbHandle, CertDbRef) of
+                          {unknown_ca, _} = Result ->
+                              handle_incomplete_chain(Chain, 
+                                                      PartialChainHandler, 
+                                                      Result,
+                                                      CertDbHandle, CertDbRef);
+                          Result ->
+                              Result
+                      end
+              end, Paths).
 %%--------------------------------------------------------------------
--spec certificate_chain(undefined | binary() | #'OTPCertificate'{} , db_handle(), certdb_ref()) ->
-			  {error, no_cert} | {ok, #'OTPCertificate'{} | undefined, [der_cert()]}.
+-spec certificate_chain(undefined | binary() | #'OTPCertificate'{} , db_handle(), certdb_ref() | {extracted, list()}) ->
+			  {error, no_cert} | {ok, der_cert() | undefined, [der_cert()]}.
 %%
 %% Description: Return the certificate chain to send to peer.
 %%--------------------------------------------------------------------
@@ -104,7 +104,7 @@ certificate_chain(OwnCert, CertDbHandle, CertsDbRef) ->
 
 %%--------------------------------------------------------------------
 -spec certificate_chain(undefined | binary() | #'OTPCertificate'{} , db_handle(), certdb_ref() | {extracted, list()}, [der_cert()]) ->
-			  {error, no_cert} | {ok, #'OTPCertificate'{} | undefined, [der_cert()]}.
+			  {error, no_cert} | {ok, der_cert() | undefined, [der_cert()]}.
 %%
 %% Description: Create certificate chain with certs from 
 %%--------------------------------------------------------------------
@@ -141,7 +141,7 @@ file_to_crls(File, DbHandle) ->
 %% Description:  Validates ssl/tls specific extensions
 %%--------------------------------------------------------------------
 validate(_,{extension, #'Extension'{extnID = ?'id-ce-extKeyUsage',
-				    extnValue = KeyUse}}, UserState = {Role, _,_, _, _, _}) ->
+				    extnValue = KeyUse}}, UserState = #{role := Role}) ->
     case is_valid_extkey_usage(KeyUse, Role) of
 	true ->
 	    {valid, UserState};
@@ -152,12 +152,28 @@ validate(_, {extension, _}, UserState) ->
     {unknown, UserState};
 validate(_, {bad_cert, _} = Reason, _) ->
     {fail, Reason};
-validate(_, valid, UserState) ->
-    {valid, UserState};
-validate(Cert, valid_peer, UserState = {client, _,_, {Hostname, Customize}, _, _}) when Hostname =/= disable ->
-    verify_hostname(Hostname, Customize, Cert, UserState);  
-validate(_, valid_peer, UserState) ->    
-   {valid, UserState}.
+validate(Cert, valid, UserState) ->
+    case verify_sign(Cert, UserState) of
+        true ->
+            case maps:get(cert_ext, UserState, undefined) of
+                undefined ->
+                    {valid, UserState};
+                _ ->
+                    verify_cert_extensions(Cert, UserState)
+            end;
+        false ->
+            {fail, {bad_cert, invalid_signature}}
+    end;
+validate(Cert, valid_peer, UserState = #{role := client, server_name := Hostname, 
+                                         customize_hostname_check := Customize}) when Hostname =/= disable ->
+    case verify_hostname(Hostname, Customize, Cert, UserState) of
+        {valid, UserState} ->
+            validate(Cert, valid, UserState);
+        Error ->
+            Error
+    end;
+validate(Cert, valid_peer, UserState) ->    
+    validate(Cert, valid, UserState).
 
 %%--------------------------------------------------------------------
 -spec is_valid_key_usage(list(), term()) -> boolean().
@@ -318,6 +334,12 @@ public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorith
 									parameters = Params},
 				      subjectPublicKey = Point}) ->
     {Point, Params};
+public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorithm = ?'id-Ed25519'},
+				      subjectPublicKey = Point}) ->
+    {Point, {namedCurve, ?'id-Ed25519'}};
+public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorithm = ?'id-Ed448'},
+				      subjectPublicKey = Point}) ->
+    {Point, {namedCurve, ?'id-Ed448'}};
 public_key(#'OTPSubjectPublicKeyInfo'{algorithm = #'PublicKeyAlgorithm'{algorithm = ?'rsaEncryption'}, 
 				      subjectPublicKey = Key}) ->
     Key;
@@ -343,31 +365,6 @@ other_issuer(OtpCert, BinCert, CertDbHandle, CertDbRef) ->
 	    end
     end.
 
-handle_path({BinCert, OTPCert}, Path, PartialChainHandler) ->
-    case public_key:pkix_is_self_signed(OTPCert) of
-	true ->
-	    {BinCert, lists:delete(BinCert, Path)};
-	false ->
-	   handle_incomplete_chain(Path, PartialChainHandler)
-    end.
-
-handle_incomplete_chain(Chain, Fun) ->
-    case catch Fun(Chain) of
-	{trusted_ca, DerCert} ->
-	    new_trusteded_chain(DerCert, Chain);
-	unknown_ca = Error ->
-	    {Error, Chain};
-	_  ->
-	    {unknown_ca, Chain}
-    end.
-
-new_trusteded_chain(DerCert, [DerCert | Chain]) ->
-    {DerCert, Chain};
-new_trusteded_chain(DerCert, [_ | Rest]) ->
-    new_trusteded_chain(DerCert, Rest);
-new_trusteded_chain(_, []) ->
-    {unknown_ca, []}.
-
 verify_hostname({fallback, Hostname}, Customize, Cert, UserState) when is_list(Hostname) ->
     case public_key:pkix_verify_hostname(Cert, [{dns_id, Hostname}], Customize) of
         true ->
@@ -396,3 +393,216 @@ verify_hostname(Hostname, Customize, Cert, UserState) ->
         false ->
             {fail, {bad_cert, hostname_check_failed}}
     end.
+
+verify_cert_extensions(Cert, #{cert_ext := CertExts} =  UserState) ->
+    Id = public_key:pkix_subject_id(Cert),
+    Extensions = maps:get(Id, CertExts, []),
+    verify_cert_extensions(Cert, UserState, Extensions, #{}).
+
+verify_cert_extensions(Cert, UserState, [], _) ->
+    {valid, UserState#{issuer => Cert}};
+verify_cert_extensions(Cert, #{ocsp_responder_certs := ResponderCerts,
+                               ocsp_state := OscpState,
+                               issuer := Issuer} = UserState, [#certificate_status{response = OcspResponsDer} | Exts], Context) ->
+    #{ocsp_nonce := Nonce} = OscpState,
+    case public_key:pkix_ocsp_validate(Cert, Issuer, OcspResponsDer, ResponderCerts, Nonce) of
+        valid ->
+            verify_cert_extensions(Cert, UserState, Exts, Context);
+        {bad_cert, _} = Status ->
+            {fail, Status}
+    end;
+verify_cert_extensions(Cert, UserState, [_|Exts], Context) ->
+    %% Skip unknow extensions!
+    verify_cert_extensions(Cert, UserState, Exts, Context).
+
+verify_sign(_, #{version := {_, Minor}}) when Minor < 3 ->
+    %% This verification is not applicable pre TLS-1.2 
+    true; 
+verify_sign(Cert, #{signature_algs := SignAlgs,
+                    signature_algs_cert := undefined}) ->
+    is_supported_signature_algorithm(Cert, SignAlgs); 
+verify_sign(Cert, #{signature_algs_cert := SignAlgs}) ->
+    is_supported_signature_algorithm(Cert, SignAlgs).
+
+is_supported_signature_algorithm(#'OTPCertificate'{signatureAlgorithm = 
+                                                       #'SignatureAlgorithm'{algorithm = ?'id-dsa-with-sha1'}}, [{_,_}|_] = SignAlgs) ->   
+    lists:member({sha, dsa}, SignAlgs);
+is_supported_signature_algorithm(#'OTPCertificate'{signatureAlgorithm = SignAlg}, [{_,_}|_] = SignAlgs) ->   
+    Scheme = ssl_cipher:signature_algorithm_to_scheme(SignAlg),
+    {Hash, Sign, _ } = ssl_cipher:scheme_to_components(Scheme),
+    lists:member({pre_1_3_hash(Hash), pre_1_3_sign(Sign)}, SignAlgs);
+is_supported_signature_algorithm(#'OTPCertificate'{signatureAlgorithm = SignAlg}, SignAlgs) ->   
+    Scheme = ssl_cipher:signature_algorithm_to_scheme(SignAlg),
+    lists:member(Scheme, SignAlgs).
+
+pre_1_3_sign(rsa_pkcs1) ->
+    rsa;
+pre_1_3_sign(Other) ->
+    Other.
+pre_1_3_hash(sha1) ->
+    sha;
+pre_1_3_hash(Hash) ->
+    Hash.
+
+paths(Chain, CertDbHandle) ->
+    paths(Chain, Chain, CertDbHandle, []).
+
+paths([Root], _, _, Path) ->
+    [[Root | Path]];
+paths([Cert1, Cert2 | Rest], Chain, CertDbHandle, Path) ->
+    case public_key:pkix_is_issuer(Cert1, Cert2) of
+        true ->
+            %% Chain orded so far
+            paths([Cert2 | Rest], Chain, CertDbHandle, [Cert1 | Path]);
+        false ->
+            %% Chain is unorded and/or contains extraneous certificates
+            unorded_or_extraneous(Chain, CertDbHandle)
+    end.
+
+unorded_or_extraneous([Peer | UnorderedChain], CertDbHandle) ->
+    ChainCandidates = extraneous_chains(UnorderedChain),
+    lists:map(fun(Candidate) ->
+                      path_candidate(Peer, Candidate, CertDbHandle)
+              end,
+              ChainCandidates).
+
+path_candidate(Peer, ChainCandidateCAs, CertDbHandle) ->
+    {ok,  ExtractedCerts} = ssl_pkix_db:extract_trusted_certs({der, ChainCandidateCAs}),
+    %% certificate_chain/4 will make sure the chain is ordered
+    case certificate_chain(Peer, CertDbHandle, ExtractedCerts, []) of
+        {ok, undefined, Chain} ->
+            lists:reverse(Chain);
+        {ok, Root, Chain} ->
+            [Root | lists:reverse(Chain)]
+    end.
+
+handle_partial_chain([IssuerCert| Rest] = Path, PartialChainHandler, CertDbHandle, CertDbRef) ->
+    case public_key:pkix_is_self_signed(IssuerCert) of
+        true -> %% IssuerCert = ROOT (That is ROOT was included in chain)
+            {ok, {SerialNr, IssuerId}} = public_key:pkix_issuer_id(IssuerCert, self),
+            case ssl_manager:lookup_trusted_cert(CertDbHandle, CertDbRef, SerialNr, IssuerId) of
+                {ok, {IssuerCert, _}} -> %% Match sent ROOT to trusted ROOT 
+                    maybe_shorten_path(Path, PartialChainHandler, {IssuerCert, Rest});
+                {ok, _} -> %% Did not match trusted ROOT
+                    maybe_shorten_path(Path, PartialChainHandler, {invalid_issuer, Path});
+                _ ->
+                    maybe_shorten_path(Path, PartialChainHandler, {unknown_ca, Path})
+            end;
+        false ->
+            OTPCert = public_key:pkix_decode_cert(IssuerCert, otp),
+            case other_issuer(OTPCert, IssuerCert, CertDbHandle, CertDbRef) of
+                {other, {SerialNr, IssuerId}} ->
+                    case ssl_manager:lookup_trusted_cert(CertDbHandle, CertDbRef, SerialNr, IssuerId) of
+                        {ok, {NewIssuerCert, _}} ->  
+                            case public_key:pkix_is_self_signed(NewIssuerCert) of
+                                true -> %% NewIssuerCert is a trusted ROOT cert
+                                    maybe_shorten_path([NewIssuerCert | Path], PartialChainHandler, {NewIssuerCert, Path});
+                                false ->
+                                    maybe_shorten_path([NewIssuerCert | Path], PartialChainHandler, 
+                                                       {unknown_ca, [NewIssuerCert | Path]})
+                            end;
+                        _ ->
+                            maybe_shorten_path(Path, PartialChainHandler, {unknown_ca, Path})
+                    end;
+                {error, issuer_not_found} ->
+                    maybe_shorten_path(Path, PartialChainHandler, {unknown_ca, Path})
+            end
+    end. 
+
+maybe_shorten_path(Path, PartialChainHandler, Default) ->
+    %% This function might shorthen the
+    %% certificate path to be validated with
+    %% public_key:pkix_path_validation by letting
+    %% the user put its trust in an intermidate cert
+    %% from the certifcate chain sent by the peer.
+    try PartialChainHandler(Path) of
+        {trusted_ca, Root} ->
+            new_trusteded_path(Root, Path, Default);
+        unknown_ca ->
+            Default
+    catch _:_ ->
+            Default
+    end.
+
+new_trusteded_path(DerCert, [DerCert | Chain], _) ->
+    {DerCert, Chain};
+new_trusteded_path(DerCert, [_ | Rest], Default) ->
+    new_trusteded_path(DerCert, Rest, Default);
+new_trusteded_path(_, [], Default) ->
+    %% User did not pick a cert present 
+    %% in the cert chain so ignore
+    Default.
+
+handle_incomplete_chain([PeerCert| _] = Chain0, PartialChainHandler, Default, CertDbHandle, CertDbRef) ->
+    %% We received an incomplete chain, that is not all certs expected to be present are present. 
+    %% See if we have the certificates to rebuild it. 
+    case certificate_chain(PeerCert, CertDbHandle, CertDbRef) of
+        {ok, _, [PeerCert | _] = Chain} when Chain =/= Chain0 -> %% Chain candidate found          
+            case lists:prefix(Chain0, Chain) of 
+                true ->
+                    handle_partial_chain(lists:reverse(Chain), PartialChainHandler, CertDbHandle, CertDbRef);
+                false ->
+                    Default
+            end;
+        _  ->
+            Default
+    end.
+
+extraneous_chains(Certs) ->
+    %% If some certs claim to be the same cert that is have the same
+    %% subject field we should create a list of possible chain certs
+    %% for each such cert. Only one chain, if any, should be
+    %% verifiable using available ROOT certs.
+    Subjects = [{subject(Cert), Cert} || Cert <- Certs],
+    Duplicates = find_duplicates(Subjects),
+    %% Number of certs with duplicates (same subject) has been limited
+    %% to 4 and the maximum number of combinations is limited to 16.
+    build_candidates(Duplicates, 4, 16).
+
+build_candidates(Map, Duplicates, Combinations) ->
+    Subjects = maps:keys(Map),
+    build_candidates(Subjects, Map, Duplicates, 1, Combinations, []).
+%%
+build_candidates([], _, _, _, _, Acc) ->
+    Acc;
+build_candidates([H|T], Map, Duplicates, Combinations, Max, Acc0) ->
+    case maps:get(H, Map) of
+	{Certs, Counter} when Counter > 1 andalso
+                              Duplicates > 0 andalso
+                              Counter * Combinations =< Max ->
+	    case Acc0 of
+		[] ->
+		    Acc = [[Cert] || Cert <- Certs],
+		    build_candidates(T, Map, Duplicates - 1, Combinations * Counter, Max, Acc);
+		_Else ->
+		    Acc = [[Cert|L] || Cert <- Certs, L <- Acc0],
+		    build_candidates(T, Map, Duplicates - 1, Combinations * Counter, Max, Acc)
+		end;
+	{[Cert|_Throw], _Counter} ->
+	    case Acc0 of
+		[] ->
+		    Acc = [[Cert]],
+		    build_candidates(T, Map, Duplicates, Combinations, Max, Acc);
+		_Else ->
+		    Acc = [[Cert|L] || L <- Acc0],
+		    build_candidates(T, Map, Duplicates, Combinations, Max, Acc)
+	    end
+    end.
+
+find_duplicates(Chain) ->
+    find_duplicates(Chain, #{}).
+%%
+find_duplicates([], Acc) ->
+    Acc;
+find_duplicates([{Subject, Cert}|T], Acc) ->
+    case maps:get(Subject, Acc, none) of
+	none ->
+	    find_duplicates(T, Acc#{Subject => {[Cert], 1}});
+	{Certs, Counter} ->
+	    find_duplicates(T, Acc#{Subject => {[Cert|Certs], Counter + 1}})
+    end.
+
+subject(Cert) ->
+    {_Serial,Subject} = public_key:pkix_subject_id(Cert),
+    Subject.
+
